#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <openssl/evp.h>
#include <openssl/pem.h>
#include <openssl/rsa.h>
#include <openssl/err.h>

#include <arpa/inet.h> /* für htonl() */

int verschluesseln(FILE *rsa_pubkey_file, FILE *in, FILE *out)
{
	int retval = EXIT_FAILURE;
	/* EVP-Kontext und Schlüssel */
	EVP_CIPHER_CTX ctx;
	RSA *rsa_pubkey = NULL;
	EVP_PKEY *pubkey = EVP_PKEY_new();
	/* Ein- und Ausgabebuffer */
	unsigned char buffer[1024];
	unsigned char buffer_out[1024 + EVP_MAX_IV_LENGTH];
	/* Größe der aus der Eingabedatei gelesenen Daten */
	size_t len;
	/* Größe der verschlüsselten Daten */
	int len_out;
	/* Session Key: Länge und Pointer */
	int sesskeylen;
	uint32_t sesskeylen_nl;
	unsigned char *sesskey = NULL;
	/* Session Key */
	unsigned char iv[EVP_MAX_IV_LENGTH];

	/* Lese Public Key des Empfängers aus Datei */
	if(!PEM_read_RSA_PUBKEY(rsa_pubkey_file, &rsa_pubkey, NULL, NULL))
	{
		fprintf(stderr, "Fehler beim Laden der Datei für den Public Key.\n");
		ERR_print_errors_fp(stderr);
		return EXIT_FAILURE;
	}

	/* Den öffentliche Key an EVP übergeben */
	if(!EVP_PKEY_assign_RSA(pubkey, rsa_pubkey))
	{
		fprintf(stderr, "EVP_PKEY_assign_RSA schlug fehl.\n");
		return EXIT_FAILURE;
	}

	/* EVP Chiffre-Kontext initialisieren */
	EVP_CIPHER_CTX_init(&ctx);
	sesskey = malloc(EVP_PKEY_size(pubkey));

	/* Initialisiere EVP seal, Schlüssel für AES generieren */
	if(!EVP_SealInit(&ctx, EVP_aes_256_cbc(), &sesskey, &sesskeylen, iv, &pubkey, 1))
	{
		fprintf(stderr, "EVP_SealInit schlug fehl.\n");
		goto out_free;
	}

	/* Ändere Byte-Order des Schlüssellängenwertes in Netzwerk-Format */
	sesskeylen_nl = htonl(sesskeylen);

	/* Schreibe Schlüssellänge, Key und Initialisierungsvektor in Datei */
	if( (fwrite(&sesskeylen_nl, sizeof(sesskeylen_nl), 1, out)        != 1) ||
	    (fwrite(sesskey, sesskeylen, 1, out)                          != 1) ||
	    (fwrite(iv, EVP_CIPHER_iv_length(EVP_aes_256_cbc()), 1, out)  != 1) )
	{
		fprintf(stderr, "Fehler beim Schreiben in Ausgabestream/-datei.\n");
		goto out_free;
	}

	/* Eingabestream durchschleifen */
	while( (len = fread(buffer, 1, sizeof(buffer), in)) > 0 )
	{
		/* Verschlüsseln des Bufferinhalts */
		if(!EVP_SealUpdate(&ctx, buffer_out, &len_out, buffer, len))
		{
			fprintf(stderr, "EVP_SealUpdate schlug fehl \n");
			goto out_free;
		}

		/* Schreiben der verschlüsselten Daten in Ausgabestream */
		if(fwrite(buffer_out, len_out, 1, out) != 1)
		{
			fprintf(stderr, "Fehler beim Schreiben in Ausgabestream/-datei.\n");
			goto out_free;
		}
	}

	/* Prüfen, ob Read-Schleife wegen Fehler abbrach */
	if(ferror(in))
	{
		fprintf(stderr, "Fehler beim lesen des Eingabestreams/-datei.\n");
		goto out_free;
	}

	/* Übergebliebene Daten aus Block verschlüsseln... */
	if(!EVP_SealFinal(&ctx, buffer_out, &len_out))
	{
		fprintf(stderr, "EVP_SealFinal schlug fehl.\n");
		goto out_free;
	}

	/* ... und speichern */
	if(fwrite(buffer_out, len_out, 1, out) != 1)
	{
		fprintf(stderr, "Fehler beim Schreiben in Ausgabestream/-datei.\n");
		goto out_free;
	}

	retval = EXIT_SUCCESS;

out_free:
	EVP_PKEY_free(pubkey);
	free(sesskey);

	return retval;
}


int entschluesseln(FILE *rsa_privkey_file, FILE *in, FILE *out)
{
	int retval = EXIT_FAILURE;
	RSA *rsa_privkey = NULL;
	EVP_PKEY *privkey = EVP_PKEY_new();
	EVP_CIPHER_CTX ctx;
	unsigned char buffer[1024];
	unsigned char buffer_out[1024 + EVP_MAX_IV_LENGTH];
	size_t len;
	int len_out;
	unsigned char *sesskey;
	unsigned int sesskeylen;
	uint32_t sesskeylen_nl;
	unsigned char iv[EVP_MAX_IV_LENGTH];

	/* Lade privaten RSA Key aus Datei */
	if(!PEM_read_RSAPrivateKey(rsa_privkey_file, &rsa_privkey, NULL, NULL))
	{
		fprintf(stderr, "Private Key kann nicht gelesen werden.\n");
		ERR_print_errors_fp(stderr);
		return EXIT_FAILURE;
	}

	/* Übergebe privaten RSA Key an EVP */
	if(!EVP_PKEY_assign_RSA(privkey, rsa_privkey))
	{
		fprintf(stderr, "EVP_PKEY_assign_RSA schlug fehl.\n");
		return EXIT_FAILURE;
	}

	/* EVP Chiffre-Kontext initialisieren */
	EVP_CIPHER_CTX_init(&ctx);
	sesskey = malloc(EVP_PKEY_size(privkey));

	/* Lese die Länge des Session Keys  */
	if(fread(&sesskeylen_nl, sizeof(sesskeylen_nl), 1, in) != 1)
	{
		fprintf(stderr, "Lesen des Eingabestreams/-datei fehlgeschlagen.\n");
		goto out_free;
	}

	/* Ändere Byteorder */
	sesskeylen = ntohl(sesskeylen_nl);

	/* Sicherstellen, dass der Session Key nicht länger als der RSA-Key ist  */
	if(sesskeylen > EVP_PKEY_size(privkey))
	{
		fprintf(stderr, "Fehler: Session Key ist länger als RSA Key (%u > %d)\n", 
		                sesskeylen, EVP_PKEY_size(privkey));
		goto out_free;
	}

	/* Lesen des verschlüsselten AES-Key aus der Datei */
	if(fread(sesskey, sesskeylen, 1, in) != 1)
	{
		fprintf(stderr, "Lesen des Eingabestreams/-datei fehlgeschlagen.\n");
		goto out_free;
	}

	/* Lesen des Initialisierungsvektors */
	if(fread(iv, EVP_CIPHER_iv_length(EVP_aes_256_cbc()), 1, in) != 1)
	{
		fprintf(stderr, "Lesen des Eingabestreams/-datei fehlgeschlagen.\n");
		goto out_free;
	}

	/* Initialisiere EVP */
	if(!EVP_OpenInit(&ctx, EVP_aes_256_cbc(), sesskey, sesskeylen, iv, privkey))
	{
		fprintf(stderr, "EVP_OpenInit fehlgeschlagen.\n");
		goto out_free;
	}

	/* Jeweils die Bufferlänge einlesen und entschlüsseln */
	while( (len = fread(buffer, 1, sizeof buffer, in)) > 0 )
	{
		/* Daten entschlüsseln */
		if (!EVP_OpenUpdate(&ctx, buffer_out, &len_out, buffer, len))
		{
			fprintf(stderr, "EVP_OpenUpdate fehlgeschlagen.\n");
			goto out_free;
		}

		/* Entschlüsselte Daten in Datei schreiben */
		if (fwrite(buffer_out, len_out, 1, out) != 1)
		{
			fprintf(stderr, "Fehler beim Schreiben in Ausgabestream/-datei.\n");
			goto out_free;
		}
	}

	/* Prüfen, ob Read-Schleife wegen Fehler abbrach */
	if (ferror(in))
	{
		fprintf(stderr, "Lesen des Eingabestreams/-datei fehlgeschlagen.\n");
		goto out_free;
	}

	/* Übergebliebene Daten aus Block entschlüsseln... */
	if (!EVP_OpenFinal(&ctx, buffer_out, &len_out))
	{
		fprintf(stderr, "EVP_SealFinal fehlgeschlagen.\n");
		goto out_free;
	}

	/* ... und speichern */
	if (fwrite(buffer_out, len_out, 1, out) != 1)
	{
		fprintf(stderr, "Fehler beim Schreiben in Ausgabestream/-datei.\n");
		goto out_free;
	}

	retval = EXIT_SUCCESS;

out_free:
	EVP_PKEY_free(privkey);
	free(sesskey);

	return retval;
}


int main(int argc, char *argv[])
{
	FILE *rsa_key_file, *in, *out;
	int retval;

	/* Prüfe Anzahl der Parameter */
	if (argc < 5)
	{
		fprintf(stderr, "Usage: %s <mode> <PEM RSA Public Key File> <in> <out>\n", argv[0]);
		fprintf(stderr, "Modes: encrypt, decrypt\n");
		exit(EXIT_FAILURE);
	}

	/* Öffne Dateien/Streams (alle Binär) */
	rsa_key_file = fopen(argv[2], "rb");
	if (!rsa_key_file)
	{
		fprintf(stderr, "Fehler beim Laden der PEM RSA Schlüsseldatei.\n");
		exit(EXIT_FAILURE);
	}
	
	in = fopen(argv[3], "rb");
	if(!in)
	{
		fprintf(stderr, "Fehler beim Öffnen der Eingabedatei: %s\n", argv[3]);
		fclose(rsa_key_file);
		exit(EXIT_FAILURE);
	}
	
	out = fopen(argv[4], "wb");
	if(!out)
	{
		fprintf(stderr, "Fehler beim Öffnen der Ausgabedatei: %s\n", argv[4]);
		fclose(in);
		fclose(rsa_key_file);
		exit(EXIT_FAILURE);
	}
	

	if( strcmp(argv[1], "encrypt") == 0 )
	{
		retval = verschluesseln(rsa_key_file, in, out);
	}
	else if( strcmp(argv[1], "decrypt") == 0 )
	{
		retval = entschluesseln(rsa_key_file, in, out);
	}
	else
	{
		fprintf(stderr, "Invalid mode: %s\n", argv[1]);
		retval = EXIT_FAILURE;
	}

	fclose(in);
	fclose(out);
	fclose(rsa_key_file);

	exit(retval);
}
